import flask
import pytest
import flask_sqlalchemy as fsa
import flask_signalbus as fsb
from mock import Mock


class SignalBusAlchemy(fsb.SignalBusMixin, fsa.SQLAlchemy):
    pass


@pytest.fixture
def app(request):
    app = flask.Flask(request.module.__name__)
    app.testing = True
    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///:memory:'
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    return app


@pytest.fixture(params=['direct', 'mixin_bound', 'mixin_init_app'])
def db(app, request):
    if request.param == 'direct':
        db = fsa.SQLAlchemy(app)
        db.signalbus = fsb.SignalBus(db)
    elif request.param == 'mixin_bound':
        db = SignalBusAlchemy(app)
    elif request.param == 'mixin_init_app':
        db = SignalBusAlchemy()
        db.init_app(app)
        db.app = app

    assert hasattr(db, 'signalbus')
    assert db.get_app()
    return db


@pytest.fixture
def signalbus(app, db):
    return db.signalbus


@pytest.fixture
def signalbus_with_pending_signal(app, db, signalbus, Signal):
    signalbus.autoflush = False
    sig = Signal(name='signal', value='1')
    db.session.add(sig)
    db.session.commit()
    signalbus.autoflush = True
    return signalbus


@pytest.fixture
def signalbus_with_pending_error(app, db, signalbus, Signal):
    signalbus.autoflush = False
    sig = Signal(name='error', value='1')
    db.session.add(sig)
    db.session.commit()
    signalbus.autoflush = True
    return signalbus


@pytest.fixture
def send_mock():
    return Mock()


@pytest.fixture
def Signal(db, send_mock):
    class Signal(db.Model):
        __tablename__ = 'test_signal'
        id = db.Column(db.Integer, primary_key=True)
        name = db.Column(db.String(60))
        value = db.Column(db.String(60))

        def send_signalbus_message(self):
            send_mock(self.id, self.name, self.value)
            if self.name == 'error':
                raise ValueError(self.value)

    db.create_all()
    yield Signal
    db.drop_all()


@pytest.fixture
def NonSignal(db):
    class NonSignal(db.Model):
        __tablename__ = 'test_non_signal'
        id = db.Column(db.Integer, primary_key=True)

    db.create_all()
    yield NonSignal
    db.drop_all()



def test_flush_empty(app, signalbus):
    runner = app.test_cli_runner()
    result = runner.invoke(args=['signalbus', 'flush'])
    assert not result.output


def test_flush_pending(app, signalbus_with_pending_signal):
    runner = app.test_cli_runner()
    result = runner.invoke(args=['signalbus', 'flush'])
    assert '1' in result.output
    assert 'processed' in result.output


def test_flush_error(app, signalbus_with_pending_error):
    runner = app.test_cli_runner()
    result = runner.invoke(args=['signalbus', 'flush'])
    assert isinstance(result.exception, ValueError)


def test_show_signals(app, signalbus_with_pending_signal):
    runner = app.test_cli_runner()
    result = runner.invoke(args=['signalbus', 'signals'])
    assert 'Signal' == result.output.strip()


def test_show_pending(app, signalbus_with_pending_signal):
    runner = app.test_cli_runner()
    result = runner.invoke(args=['signalbus', 'pending'])
    assert 'Signal' in result.output
    assert 'Total pending: 1' in result.output



"""
Adds to Flask-SQLAlchemy the capability to atomically send
messages (signals) over a message bus.
"""

import time
import logging
import random
from functools import wraps
from sqlalchemy import event, inspect
from sqlalchemy.exc import DBAPIError
from sqlalchemy.sql.expression import and_
from . import cli

ERROR_CODE_ATTRS = ['pgcode', 'sqlstate']
DEADLOCK_ERROR_CODES = ['40001', '40P01']
SIGNALS_TO_FLUSH_SESSION_INFO_KEY = 'flask_signalbus__signals_to_flush'


def get_db_error_code(exception):
    """Return 5-character SQLSTATE code, or '' if not available.

    Currently `psycopg2`, `psycopg2cffi`, and `MySQL Connector` are supported.
    """

    for attr in ERROR_CODE_ATTRS:
        error_code = getattr(exception, attr, '')
        if error_code:
            break
    return error_code


def retry_on_deadlock(session, retries=6, min_wait=0.1, max_wait=10.0):
    """Return function decorator that executes the function again in case of a deadlock."""

    def decorator(action):
        """Function decorator that retries `action` in case of a deadlock."""

        @wraps(action)
        def f(*args, **kwargs):
            num_failures = 0
            while True:
                try:
                    return action(*args, **kwargs)
                except DBAPIError as e:
                    num_failures += 1
                    if num_failures > retries or get_db_error_code(e.orig) not in DEADLOCK_ERROR_CODES:
                        raise
                session.rollback()
                wait_seconds = min(max_wait, min_wait * 2 ** (num_failures - 1))
                time.sleep(wait_seconds)

        return f

    return decorator


def is_serializaion_error(e):
    return isinstance(e, DBAPIError) and get_db_error_code(e.orig) in DEADLOCK_ERROR_CODES


class SignalBusMixin(object):
    """A **mixin class** that can be used to extend the
    `flask_sqlalchemy.SQLAlchemy` class to handle signals.

    For example::

        from flask import Flask
        from flask_sqlalchemy import SQLAlchemy
        from flask_signalbus import SignalBusMixin

        class CustomSQLAlchemy(SignalBusMixin, SQLAlchemy):
            pass

        app = Flask(__name__)
        db = CustomSQLAlchemy(app)
        db.signalbus.flush()

    """

    def init_app(self, app, *args, **kwargs):
        super(SignalBusMixin, self).init_app(app, *args, **kwargs)
        self.signalbus._init_app(app)

    @property
    def signalbus(self):
        """The associated `SignalBus` object."""

        try:
            signalbus = self.__signalbus
        except AttributeError:
            signalbus = self.__signalbus = SignalBus(self, init_app=False)
        return signalbus


class SignalBus(object):
    """Instances of this class automatically send signal messages that
    have been recorded in the SQL database, over a message
    bus. Normally, the sending of the recorded messages (if there are
    any) is done after each transaction commit, but it also can be
    triggered explicitly by a command.

    :param db: The `flask_sqlalchemy.SQLAlchemy` instance

    For example::

        from flask_sqlalchemy import SQLAlchemy
        from flask_signalbus import SignalBus

        app = Flask(__name__)
        db = SQLAlchemy(app)
        signalbus = SignalBus(db)
        signalbus.flush()

    """

    def __init__(self, db, init_app=True):
        self.db = db
        self.signal_session = self.db.create_scoped_session({'expire_on_commit': False})
        self.logger = logging.getLogger(__name__)
        self._autoflush = True
        retry = retry_on_deadlock(self.signal_session, retries=10, max_wait=1.0)
        self._flush_signals_with_retry = retry(self._flush_signals)
        event.listen(self.db.session, 'transient_to_pending', self._transient_to_pending_handler)
        event.listen(self.db.session, 'after_commit', self._after_commit_handler)
        if init_app:
            if db.app is None:
                raise RuntimeError(
                    'No application found. The SQLAlchemy instance passed to'
                    ' SignalBus should be constructed with an application.'
                )
            self._init_app(db.app)

    @property
    def autoflush(self):
        """Setting this property to `False` instructs the `SignalBus` instance
        to not automatically flush pending signals after each
        transaction commit. Setting it back to `True` restores the
        default behavior.

        """

        return self._autoflush

    @autoflush.setter
    def autoflush(self, value):
        self._autoflush = bool(value)

    def get_signal_models(self):
        """Return all signal types in a list.

        :rtype: list(`signal-model`)

        """

        base = self.db.Model
        return [
            cls for cls in base._decl_class_registry.values() if (
                isinstance(cls, type) and
                issubclass(cls, base) and
                hasattr(cls, 'send_signalbus_message')
            )
        ]

    def flush(self, model=None):
        """Send all pending signals over the message bus.

        :param model: If passed, flushes only signals of the specified type.
        :type model: `signal-model` or `None`
        :return: The total number of signals that have been sent

        """

        models_to_flush = [model] if model else self.get_signal_models()
        try:
            return sum(self._flush_signals_with_retry(m) for m in models_to_flush)
        finally:
            self.signal_session.remove()

    def _init_app(self, app):
        if not hasattr(app, 'extensions'):
            app.extensions = {}
        if app.extensions.get('signalbus') not in [None, self]:
            raise RuntimeError('Can not attach more than one SignalBus to one application.')
        app.extensions['signalbus'] = self
        app.cli.add_command(cli.signalbus)

    def _transient_to_pending_handler(self, session, instance):
        model = type(instance)
        if hasattr(model, 'send_signalbus_message'):
            signals_to_flush = session.info.setdefault(SIGNALS_TO_FLUSH_SESSION_INFO_KEY, set())
            signals_to_flush.add(instance)

    def _after_commit_handler(self, session):
        signals_to_flush = session.info.setdefault(SIGNALS_TO_FLUSH_SESSION_INFO_KEY, set())
        if self.autoflush:
            for signal in signals_to_flush:
                model = type(signal)
                m = inspect(model)
                pk_attrs = [m.get_property_by_column(c).class_attribute for c in m.primary_key]
                pk_values = m.primary_key_from_instance(signal)
                clause = and_(*[attr == value for attr, value in zip(pk_attrs, pk_values)])
                try:
                    self.signal_session.query(model).filter(clause).delete()
                    signal.send_signalbus_message()
                except Exception as e:
                    if not is_serializaion_error(e):
                        self.logger.exception('Caught error while flushing %s.', model.__name__)
                    self.signal_session.rollback()
                    break
            self.signal_session.commit()
            self.signal_session.expire_all()
        elif signals_to_flush:
            self.logger.debug('Flushing skipped, "autoflush" is False.')
        signals_to_flush.clear()

    def _flush_signals(self, model):
        if not hasattr(model, 'send_signalbus_message'):
            raise RuntimeError(
                '{} can not be flushed because it does not have a'
                ' "send_signalbus_message" method.'
            )
        self.logger.debug('Flushing %s.', model.__name__)
        burst_count = int(getattr(model, 'signalbus_burst_count', 1))
        signal_count = 0
        signals = self.signal_session.query(model).all()
        self.signal_session.commit()
        if burst_count > 1:
            # Shuffle signals randomly to avoid systematic deadlocks.
            signals = list(signals)
            random.shuffle(signals)
        for signal in signals:
            self.signal_session.delete(signal)
            signal.send_signalbus_message()
            signal_count += 1
            if signal_count % burst_count == 0:
                self.signal_session.commit()
        self.signal_session.commit()
        self.signal_session.expire_all()
        return signal_count



import pytest
from flask_sqlalchemy import SQLAlchemy
from flask_signalbus import SignalBus
from .conftest import SignalBusAlchemy


def test_create_signalbus_alchemy(app):
    assert 'signalbus' not in app.extensions
    db = SignalBusAlchemy(app)
    assert 'signalbus' in app.extensions
    assert len(db.signalbus.get_signal_models()) == 0


def test_create_signalbus_alchemy_init_app(app):
    db = SignalBusAlchemy()
    assert 'signalbus' not in app.extensions
    db.init_app(app)
    assert 'signalbus' in app.extensions
    assert len(db.signalbus.get_signal_models()) == 0


def test_create_signalbus_directly(app):
    assert 'signalbus' not in app.extensions
    db = SQLAlchemy(app)
    signalbus = SignalBus(db)
    assert 'signalbus' in app.extensions
    assert not hasattr(db, 'signalbus')
    assert len(signalbus.get_signal_models()) == 0


def test_create_two_signalbuses_directly(app):
    db = SQLAlchemy(app)
    SignalBus(db)
    with pytest.raises(RuntimeError):
        SignalBus(db)


def test_create_signalbus_directly_no_app():
    db = SQLAlchemy()
    with pytest.raises(RuntimeError):
        SignalBus(db)


def test_flush_signal_model(app, signalbus, Signal):
    assert len(signalbus.get_signal_models()) == 1
    signalbus.flush(Signal)


def test_flush_nonsignal_model(app, signalbus, NonSignal):
    assert len(signalbus.get_signal_models()) == 0
    with pytest.raises(RuntimeError):
        signalbus.flush(NonSignal)


def test_flush_all_signal_models(app, signalbus, Signal, NonSignal):
    assert len(signalbus.get_signal_models()) == 1
    signalbus.flush()


def test_send_signal_success(db, signalbus, send_mock, Signal):
    sig = Signal(name='signal', value='1')
    db.session.add(sig)
    db.session.flush()
    sig_id = sig.id
    send_mock.assert_not_called()
    db.session.commit()
    send_mock.assert_called_once_with(sig_id, 'signal', '1')
    assert Signal.query.count() == 0


def test_send_signal_error(db, signalbus, send_mock, Signal):
    sig = Signal(name='error', value='1')
    db.session.add(sig)
    db.session.commit()
    assert send_mock.call_count == 1
    with pytest.raises(ValueError):
        signalbus.flush()
    assert send_mock.call_count == 2
    assert Signal.query.count() == 1


def test_autoflush_false(db, signalbus, send_mock, Signal):
    db.session.add(Signal(name='signal', value='1'))
    signalbus.autoflush = False
    db.session.commit()
    assert send_mock.call_count == 0
    assert Signal.query.count() == 1
    signalbus.flush()
    assert send_mock.call_count == 1
    assert Signal.query.count() == 0

    db.session.add(Signal(name='signal', value='2'))
    signalbus.autoflush = True
    db.session.commit()
    assert send_mock.call_count == 2
    assert Signal.query.count() == 0


def test_send_nonsignal_model(db, signalbus, send_mock, NonSignal):
    assert len(signalbus.get_signal_models()) == 0
    db.session.add(NonSignal())
    db.session.flush()
    db.session.commit()
    assert NonSignal.query.count() == 1



import click
from flask.cli import with_appcontext
from flask import current_app


@click.group()
def signalbus():
    """Perform SignalBus operations."""


@signalbus.command()
@with_appcontext
@click.option('-e', '--exclude', multiple=True, help="Do not flush signals with the specified name.")
@click.argument('signal_names', nargs=-1)
def flush(signal_names, exclude):
    """Send pending signals over the message bus.

    If a list of SIGNAL_NAMES is specified, flushes only those
    signals. If no SIGNAL_NAMES are specified, flushes all signals.

    """

    signalbus = current_app.extensions['signalbus']
    signal_names = set(signal_names)
    exclude = set(exclude)
    models_to_flush = signalbus.get_signal_models()
    if signal_names and exclude:
        click.echo('Warning: Specified both SIGNAL_NAMES and exclude option.')
    if signal_names:
        wrong_signal_names = signal_names - {m.__name__ for m in models_to_flush}
        models_to_flush = [m for m in models_to_flush if type(m).__name__ in signal_names]
    else:
        wrong_signal_names = exclude - {m.__name__ for m in models_to_flush}
    for name in wrong_signal_names:
        click.echo('Warning: A signal with name "{}" does not exist.'.format(name))
    models_to_flush = [m for m in models_to_flush if type(m).__name__ not in exclude]
    signal_count = 0
    for model in models_to_flush:
        signal_count += signalbus.flush()
    if signal_count == 1:
        click.echo('{} signal has been successfully processed.'.format(signal_count))
    elif signal_count > 1:
        click.echo('{} signals have been successfully processed.'.format(signal_count))


@signalbus.command()
@with_appcontext
def signals():
    """Show all signal types."""

    signalbus = current_app.extensions['signalbus']
    for signal_model in signalbus.get_signal_models():
        click.echo(signal_model.__name__)


@signalbus.command()
@with_appcontext
def pending():
    """Show the number of pending signals by signal type."""

    signalbus = current_app.extensions['signalbus']
    pending = []
    total_pending = 0
    for signal_model in signalbus.get_signal_models():
        count = signal_model.query.count()
        if count > 0:
            pending.append((count, signal_model.__name__))
        total_pending += count
    if pending:
        pending.sort()
        max_chars = len(str(pending[-1][0]))
        for n, signal_name in pending:
            click.echo('{} of type "{}"'.format(str(n).rjust(max_chars), signal_name))
    click.echo(25 * '-')
    click.echo('Total pending: {} '.format(total_pending))
